/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* 
 * Note to regenerate the Makefile and configure script:
 * aclocal
 * autoconf
 * touch NEWS README AUTHORS ChangeLog
 * automake -a
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <assert.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <wctype.h>
#include <wchar.h>
#include <locale.h>
#include <langinfo.h>

#undef HAVE_LIBBOOST_REGEX

#if defined HAVE_LIBBOOST_REGEX

#define UNICODE
#include <boost/regex.h>

#else 

#include <sys/types.h>
#include <regex.h>

#endif

/* digram digitization: avg loss of precision = 0.02 * token size */
#define DIGITIZE_DIGRAMS
/* lambda digitization: avg loss of precision = 0.02 */
#define DIGITIZE_LAMBDA
#include "dbacl.h" /* make sure this is last */

/* global variables */

long int default_max_token_bits = 16;
long int default_max_tokens = (1<<16);

Dirichlet dirichlet;

Learner learner;
Category cat[MAX_CAT];
int cat_count = 0;
int filter[MAX_CAT];
int filter_count = 0;
Regex re[MAX_RE];
int regex_count = 0;
Empirical empirical;

char *textbuf = NULL;
int textbuf_len = 0;
wchar_t *wc_textbuf = NULL;
int wc_textbuf_len = 0;
char *aux_textbuf = NULL;
int aux_textbuf_len = 0;


MBOX_State mbox;
XML_State xml;

/* for option processing */
extern char *optarg;
extern int optind, opterr, optopt;

unsigned long int options = 0;
char *title = "";

/* string hash function */
extern unsigned long int hash( unsigned char *k, 
			       unsigned long int length, 
			       unsigned long int initval);

int exit_code = 0; /* default */

/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

void usage(char **argv) {
  fprintf(stderr, 
	  "usage1: %s [-vnNR] [-T type] -c CATEGORY [-c CATEGORY]...\n", 
	  argv[0]);
  fprintf(stderr, 
	  "        [-f KEEP]... [FILE]...\n");
  fprintf(stderr, 
	  "       classifies FILE or STDIN using CATEGORY, optionally\n");
  fprintf(stderr, 
	  "       removing all lines which don't fit the category KEEP\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "usage2: %s [-vnrND] [-h size] [-T type] -l CATEGORY \n", argv[0]);
  fprintf(stderr, 
	  "        [-g regex]... [FILE]...\n");
  fprintf(stderr, 
	  "       builds a maximum entropy model from the words in FILE or STDIN\n");
  fprintf(stderr, 
	  "usage3: %s -V\n", argv[0]);
  fprintf(stderr, 
	  "       prints program version\n");
}


/***********************************************************
 * EMPIRICAL DISTRIBUTION OF TOKENS                        *
 ***********************************************************/

/* initialize global learner object */
void init_empirical() {

    /* some constants */
    empirical.max_tokens = default_max_tokens;
    empirical.max_token_bits = default_max_token_bits;
    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.cached = 0;
    empirical.stack_top = 0;

    /* allocate room for hash */
    empirical.hash = calloc(empirical.max_tokens, sizeof(h_item));
    if( !empirical.hash ) {
	fprintf(stderr, 
		"error: not enough memory? I couldn't allocate %li bytes\n", 
		(sizeof(h_item) * empirical.max_tokens));
	exit(0);
    }
}

void clear_empirical() {
    int i;

    if( empirical.cached ) {
	/* this may actually be slower than a global memset */ 
	for(i = 0; i < empirical.stack_top; i++) {
	    memset(empirical.stack[i], 0, sizeof(h_item));
	}
    } else {
	memset(empirical.hash, 0, sizeof(h_item) * empirical.max_tokens);
    }

    empirical.full_token_count = 0;
    empirical.unique_token_count = 0;
    empirical.stack_top = 0;
}

h_item *find_in_empirical(unsigned long int id) {
  register h_item *i, *loop;
  /* start at id */
  i = loop = &empirical.hash[id & (empirical.max_tokens - 1)];

  while( FILLEDP(i) ) {
    if( EQUALP(i->id,id) ) {
      return i; /* found id */
    } else {
      i++; /* not found */
      /* wrap around */
      i = (i >= &empirical.hash[empirical.max_tokens]) ? empirical.hash : i; 
      if( i == loop ) {
	return NULL; /* when hash table is full */
      }
    }
  }

  /* empty slot, so not found */

  return i; 
}

/* calculates the entropy of the empirical measure */
double empirical_entropy() {
    int i;
    double e = 0.0;

    if( empirical.cached && empirical.stack_top < MAX_TOKEN_LINE_STACK) {
	for(i = 0; i < empirical.stack_top; i++) {
	    e += ((double)empirical.stack[i]->count) * 
	      log((double)empirical.stack[i]->count);
	}
    } else {
	for(i = 0; i < empirical.max_tokens; i++) {
	    if( FILLEDP(&empirical.hash[i]) ) {
		e += ((double)empirical.hash[i].count) * 
		  log((double)empirical.hash[i].count);
	    }
	}
    }
    e = e/empirical.full_token_count - log((double)empirical.full_token_count);


    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "empirical entropy = %g\n", -e);
    }

    return -e;
}

/***********************************************************
 * CATEGORY FUNCTIONS                                      *
 ***********************************************************/

/* initialize to zero. Filename specified elsewhere */
void init_category(Category *cat) {
    char *p;

    cat->full_token_count = 0;
    cat->unique_token_count = 0;
    cat->score = 0.0;
    if( (p = strrchr(cat->filename, '/')) ) {
	cat->filename = p + 1; /* only keep basename */
    }
    cat->retype = 0;
}

/* turns purely random text into a category of its own */
void init_purely_random_text_category(Category *cat) {
    int i, j;
    double z = -log((double)ASIZE);

#if defined DIGITIZE_DIGRAMS
    short int zz = (short int)(100 * z);
#endif
    
    for(i = 1; i < ASIZE; i++) {
      for(j = 0; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
	cat->dig[i][j] = zz;
#else
	cat->dig[i][j] = z;
#endif
      }
    }
    z = -log((double)ASIZE - 1.0);
#if defined DIGITIZE_DIGRAMS
    zz = (short int)(100 * z);
#endif
    for(j = 1; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
      cat->dig[0][j] = zz;
#else
      cat->dig[0][j] = z;
#endif
    }
    /* not needed: set 00 score for completeness only */
#if defined DIGITIZE_DIGRAMS
    cat->dig[0][0] = SHRT_MIN;
#else
    cat->dig[0][0] = log(0.0);
#endif

    cat->logZ = 0.0;
    cat->hash = NULL;
}

c_item *find_in_category(Category *cat, unsigned long int id) {
    register c_item *i, *loop;

    if( cat->hash ) {
	/* start at id */
	i = loop = &cat->hash[id & (cat->max_tokens - 1)];

	while( FILLEDP(i) ) {
	    if( EQUALP(i->id,id) ) {
		return i; /* found id */
	    } else {
		i++; /* not found */
		/* wrap around */
		i = (i >= &cat->hash[cat->max_tokens]) ? cat->hash : i; 
		if( i == loop ) {
		    return NULL; /* when hash table is full */
		}
	    }
	}
	return i;
    } else {
	return NULL;
    }
}

/* for each loaded category, this calculates the score */
void score_word(char *tok) {
  int i;
  int pp, pc;
  unsigned long int id;
  char *q;
  register c_item *k;
  h_item *h;

  if( *tok ) {
    id = hash(tok, strlen(tok), 0);

    if( cat_count == 1 ) { /* calculate empirical distribution */
      h = find_in_empirical(id);
      if( h ) {
	if( FILLEDP(h) ) {
	  h->count++;
	} else if( /* !FILLEDP(i) && */
		  ((100 * empirical.unique_token_count) <
		   (HASH_FULL * empirical.max_tokens) )) {
	  /* fill the empirical hash */
	  SET(h->id,id);
	  empirical.unique_token_count++;
	  h->count++;
	}
	if( empirical.cached && 
	    (empirical.stack_top < MAX_TOKEN_LINE_STACK)) {
	  empirical.stack[empirical.stack_top++] = h;
	}
      }
    }
    empirical.full_token_count++; /* even if cat_count > 1 */

    /* now do scoring for all available categories */
    for(i = 0; i < cat_count; i++) {
      /* see if token is interesting */
      k = find_in_category(&cat[i], id);
      /* if token found, add its lambda weight */
      if( k ) {
#if defined DIGITIZE_LAMBDA
	cat[i].score += ((double)k->lam)/100.00;
#else
	cat[i].score += k->lam;
#endif
      }
   
      /* now compute the reference weight from digram model */
      q = tok;
      pp = 0;
      while( *q ) {
	pc = (unsigned char)*q;
#if defined DIGITIZE_DIGRAMS
	cat[i].score += ((double)cat[i].dig[pp][pc])/100.00;
#else
	cat[i].score += cat[i].dig[pp][pc];
#endif
	pp = pc;
	q++;
      }
      /* trailing space */
#if defined DIGITIZE_DIGRAMS
      cat[i].score += ((double)cat[i].dig[pp][0])/100.00; 
#else
      cat[i].score += cat[i].dig[pp][0]; 
#endif
      /* don't forget the normalizing constant */
      cat[i].score -= cat[i].logZ;      
    }
  }
}

void line_score_single_category() {
    double score = 0.0;

    if( *textbuf ) {
      if( options & (1<<OPTION_POSTERIOR) ) {
	score = -cat[0].score/empirical.full_token_count;
	fprintf(stdout, "%8.2f %s", score, textbuf);
      }
      if( options & (1<<OPTION_SCORES) ) {
	score = -empirical_entropy() - cat[0].score/empirical.full_token_count;
	fprintf(stdout, "%8.2f %s", score, textbuf);
      }
    }
    cat[0].score = 0.0;
    clear_empirical();
}

void line_score_multiple_categories() {
    int i, k;
    double c, cmin;

    cmin = -cat[0].score;
    k = 0;
    for(i = 1; i < cat_count; i++) {
	if(cmin > -cat[i].score) {
	    cmin = -cat[i].score;
	    k = i;
	}
    }

    if(options & (1<<OPTION_POSTERIOR) ) {
      /* compute probabilities given exclusive choices */
      c = 0.0;
      for(i = 0; i < cat_count; i++) {
	c += exp((cat[i].score + cmin));
      }
      
      if( *textbuf ) {
	for(i = 0; i < cat_count; i++) {
	  fprintf(stdout, "%s %6.2f%% ", 
		  cat[i].filename, 100 * exp((cat[i].score + cmin))/c);
	}
	fprintf(stdout, "%s", textbuf);
      }
    } else if( options & (1<<OPTION_SCORES) ) {
      /* display normalized divergence scores */
      if( *textbuf ) {
	for(i = 0; i < cat_count; i++) {
	  fprintf(stdout, "%s %6.2f ", 
		  cat[i].filename, -cat[i].score/empirical.full_token_count);
	}
	fprintf(stdout, "%s", textbuf);
      }
    } else {
      /* prune the text which doesn't fit */
      for(i = 0; i < filter_count; i++) {
	if( k == filter[i] ) {
	  fprintf(stdout, "%s", textbuf);
	  break;
	}
      }
    }
    
    /* clean up for next line */
    for(i = 0; i < cat_count; i++) {
	cat[i].score = 0.0;
    }
    empirical.full_token_count = 0;
}

void score_single_category() {
    double score = 0.0;

    if( options & (1<<OPTION_POSTERIOR) ) {
      score = -cat[0].score/empirical.full_token_count;
      fprintf(stdout, "cross_entropy: %8.2f\n", score);
    }

    if( options & (1<<OPTION_SCORES) ) {
	score = -empirical_entropy() - cat[0].score/empirical.full_token_count;
	fprintf(stdout, "KL_score: %8.2f\n", score);
    }
    exit_code = 1;
}

void score_multiple_categories() {
    int i;
    double c, cmin;

    /* find the maximum likelihood category */
    cmin = -cat[0].score;
    exit_code = 0;
    for(i = 1; i < cat_count; i++) {
	if(cmin > -cat[i].score) {
	    cmin = -cat[i].score;
	    exit_code = i;
	}
    }

    /* here we compute probabilities given exclusive choices */
    if( options & (1<<OPTION_POSTERIOR) ) {

	c = 0.0;
	for(i = 0; i < cat_count; i++) {
	    c += exp((cat[i].score + cmin));
	}

	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %5.2f%% ", 
		    cat[i].filename, 100 * exp((cat[i].score + cmin))/c);
	}
	fprintf(stdout, "\n");
    }

    /* display normalized divergence scores */
    if( options & (1<<OPTION_SCORES) ) {
	for(i = 0; i < cat_count; i++) {
	    fprintf(stdout, "%s %5.2f ", 
		    cat[i].filename, -cat[i].score/empirical.full_token_count);
	}
	fprintf(stdout, "\n");
    }

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, "%s\n", cat[exit_code].filename); 
    }

    exit_code++;
}

/***********************************************************
 * FILE MANAGEMENT FUNCTIONS                               *
 ***********************************************************/

/* used by both load and save functions */
#define MAGIC1 "# dbacl version "VERSION" category %s %s\n"
#define MAGIC2 "# model divergence: %lg logZ: %lg\n"
#define MAGIC3 "# hash_size: %hi full_token_count: %li unique_token_count: %li\n"
#define MAGIC4 "# case sensitive\n"
#define MAGIC5 "# regex %s\n"
#define MAGIC6 "#\n"
#define MAGIC7 "# wide characters\n"

/* loads a category hash */
int load_category(Category *cat) {
    long int i, j;
    int r;
    char buf[256];

    FILE *input;

    if( (input = fopen(cat->filename, "rb")) ) {
	fgets(buf, 256, input);
	if( strncmp(buf, MAGIC1, 16 + strlen(VERSION)) ) {
	    fprintf(stderr, 
		    "error: not a dbacl version "VERSION" category file [%s]\n", 
		    cat->filename);
	    return 0;
	} 
	fgets(buf, 256, input);
	sscanf(buf, MAGIC2, &cat->divergence, &cat->logZ);
	fgets(buf, 256, input);
	sscanf(buf, MAGIC3, &cat->max_token_bits, 
	       &cat->full_token_count, &cat->unique_token_count);
	cat->max_tokens = (1<<cat->max_token_bits);

	/* see if there are any regexes */
	fgets(buf, 256, input);
	while(1) {
	  if( strncmp(buf, MAGIC6, 2) == 0 ) {
	    break;
	  } else if( strncmp(buf, MAGIC5, 8) == 0 ) {
#if defined HAVE_LIBBOOST_REGEX
	    /* boost regexes accept wide characters */
	    
	    r = strlen(buf + 9);
	    re[regex_count].string = malloc((r + 1) * sizeof(wchar_t));
	    /* note: skip the trailing \n */
	    r = mbstowcs(re[regex_count].string, buf + 9, r - 1); 
	    if( r <= 0 ) {
	      fprintf(stderr, 
		      "error: couldn't convert regular expression '%s'.\n",
		      buf + 9);
	      exit(0); /* no point in going on */
	    } else {
	      re[regex_count].string[r] = L'\0';
	    }
	    /* now see if we already have this string */
	    for(r = 0; r < regex_count; r++) {
	      if( wcscmp(re[r].string, re[regex_count].string) == 0 ) {
		break;
	      }
	    }
	    if( r < regex_count ) {
	      /* we found this regex, no need to repeat it */
	      free(re[regex_count].string);
	    } else {
	      /* compile the regex (uses regcompW) */
	      if( regcomp(&re[regex_count].regex,
			   re[regex_count].string, REG_EXTENDED) != 0 ) {
		fprintf(stderr, 
			"error: could not compile regular expression '%ls'.\n", 
			re[regex_count].string);
		exit(0);
	      } else {
		regex_count++;
		if( regex_count >= MAX_RE ) { 
		  fprintf(stderr, "error: too many regular expressions\n");
		  exit(0); /* no point in going on */
		}
	      }
	    }
#else
	    /* GNU regexes use ordinary strings */
	    /* it's a regex - see if we've got it already */
	    for(r = 0; r < regex_count; r++) {
	      if( strcmp(re[r].string, buf + 9) == 0 ) {
		break;
	      }
	    }
	    if( r >= regex_count ) { /* not found */
	      /* add it to our list */
	      re[regex_count].string = strdup(buf + 9);
	      re[regex_count].string[strlen(re[regex_count].string) - 1] = 0;
	      /* and compile the regex */
	      if( regcomp(&re[regex_count].regex, 
			  re[regex_count].string, REG_EXTENDED) != 0 ) {
		fprintf(stderr, 
			"error: could not compile regular expression '%s'.\n", 
			re[regex_count].string);
		exit(0); /* not much point going on */
	      } else {
		regex_count++;
		if( regex_count >= MAX_RE ) { 
		  fprintf(stderr, "error: too many regular expressions\n");
		  exit(0); /* no point in going on */
		}
	      }
	    }

	    /* now flag the regex as taken */
	    cat->retype |= (1<<r);
#endif
	  } else if( strncmp(buf, MAGIC4, 10) == 0) {
	    /* it's the case sensitive flag */
	    options |= (1<<OPTION_CASEN);
	  } else if( strncmp(buf, MAGIC7, 5) == 0) {
	    options |= (1<<OPTION_I18N);
	  }

	  fgets(buf, 256, input); /* finished with current line, get next one */
	}

	/* if this category did not register a regex, it wants
	   the default processing, so we flag this */
	if( !cat->retype ) {
	  options |= (1<<OPTION_NOREGEX);
	}

	/* read character frequencies */
#if defined DIGITIZE_DIGRAMS
	fread(cat->dig, sizeof(short int), ASIZE * ASIZE, input);
#else
	fread(cat->dig, sizeof(float), ASIZE * ASIZE, input);
#endif
	/* allocate hash table */
	if( !(cat->hash = malloc(sizeof(c_item) * cat->max_tokens)) ) {
	    fprintf(stderr, "error: not enough memory for category %s\n", 
		    cat->filename);
	    return 0;
	}
	/* read in hash table */
	i = cat->max_tokens;
	j = 0;
	while( j < i ) {
	    j += fread(cat->hash + j, sizeof(c_item), i, input);
	}
	fclose(input);
    } else {
	fprintf(stderr, "error: cannot open file for reading %s\n", 
		cat->filename);
	return 0;
    }

    return 1;
}


/* writes the learner to a file for easily readable category */
int save_learner() {
  long int i, j;
  FILE *output;
  c_item ci;
  char buf[256];

#if defined DIGITIZE_DIGRAMS
  short int shval;
#endif

  if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, "saving category to file %s\n", learner.filename);
  }
  
  /* don't overwrite data files */
  if( (output = fopen(learner.filename, "r")) ) {
    /* output file exists already */
    fgets(buf, 256, output);
    if( strncmp(buf, MAGIC1, 10) != 0 ) {
      fprintf(stderr,
	      "error: the file %s is already used for something, "
	      "use another filename. Nothing written.\n", learner.filename);
      fclose(output);
      exit(0);
    } else {
      /* it's an existing category file */
      fclose(output);
    }
  }
  
  if( (output = fopen(learner.filename, "wb")) ) {
    /* print out standard headers */
    fprintf(output, MAGIC1, learner.filename, 
	    (options & (1<<OPTION_REFMODEL)) ? "(ref)" : "");
    fprintf(output, MAGIC2, learner.divergence, learner.logZ);
    fprintf(output, MAGIC3, learner.max_token_bits, 
	    learner.full_token_count, learner.unique_token_count);

    /* print out any regexes we might need */
    for(i = 0; i < regex_count; i++) {
#if defined HAVE_LIBBOOST_REGEX
      /* does this work? */
      fprintf(output, MAGIC5, re[i].string);
#else
      fprintf(output, MAGIC5, re[i].string);
#endif
    }
    /* this is optional too */
    if( options & (1<<OPTION_CASEN) ) {
      fprintf(output, MAGIC4);
    }
    if( options & (1<<OPTION_I18N) ) {
      fprintf(output, MAGIC7);
    }
    fprintf(output, MAGIC6); 

    /* end of readable stuff */

    /* character frequencies */
    for(i = 0; i < ASIZE; i++) {
      for(j = 0; j < ASIZE; j++) {
#if defined DIGITIZE_DIGRAMS
	if( isinf(learner.dig[i][j]) ) {
	  shval = SHRT_MIN;
	} else {
	  shval = (short int)(100 * learner.dig[i][j]);
	}
	while(fwrite(&shval, sizeof(short int), 1, output) < 1);
#else
	while(fwrite(&learner.dig[i][j], sizeof(float), 1, output) < 1);
#endif
      }
    }

    /* token/feature weights */
    for(i = 0; i < learner.max_tokens; i++) {
      /* write each element so that it's easy to read back in a c_item array */
      ci.id[0] = learner.hash[i].id[0];
      ci.id[1] = learner.hash[i].id[1];      
      ci.id[2] = learner.hash[i].id[2];
#if defined DIGITIZE_LAMBDA
      ci.lam = (short int)(learner.hash[i].lam * 100);
      if( fabs(learner.hash[i].lam) > 320 ) {
	fprintf(stderr,
		"warning: digitized an extreme lambda value - results may be unusable\n");
      } 
#else
      ci.lam = learner.hash[i].lam;
#endif
      while(fwrite(&ci, sizeof(ci), 1, output) < 1); 
    }
    fclose(output);
  } else {
    fprintf(stderr, "error: cannot open file for writing %s\n", 
	    learner.filename);
    return 0;
  }

  return 1;
}


/***********************************************************
 * LEARNER FUNCTIONS                                       *
 ***********************************************************/

l_item *find_in_learner(unsigned long int id) {
    register l_item *i, *loop;
    /* start at id */
    i = loop = &learner.hash[id & (learner.max_tokens - 1)];

    while( FILLEDP(i) ) {
	if( EQUALP(i->id,id) ) {
	    return i; /* found id */
	} else {
	    i++; /* not found */
	    /* wrap around */
	    i = (i >= &learner.hash[learner.max_tokens]) ? learner.hash : i; 
	    if( i == loop ) {
		return NULL; /* when hash table is full */
	    }
	}
    }

    /* empty slot, so not found */

    return i; 
}

/* places the token in the global hash and writes the
   token to a temporary file for later, then updates
   the digram frequencies */
void hash_word_and_learn(char *tok) {
  unsigned long int id;
  l_item *i;
  int p,q;

  if( *tok ) {

    id = hash(tok, strlen(tok), 0);
    i = find_in_learner(id);
    if( i ) {
      if( FILLEDP(i) ) {
	i->count++; 
	learner.full_token_count++;
      } else if( /* !FILLEDP(i) && */
		((100 * learner.unique_token_count) < 
		 (HASH_FULL * learner.max_tokens) )) { 
	/* fill the hash and write to file */
	SET(i->id,id);
	learner.unique_token_count++;
	learner.full_token_count++;
	i->count++; 

     	if( (options & (1<<OPTION_DEBUG)) &&
	    !(options & (1<<OPTION_NOREGEX)) ) {
     	  fprintf(stdout, "match %s\n", tok);
     	}

	fputs(tok, learner.tmp);
	fputc(-1, learner.tmp);
      }
    }

    /* now update digram frequency counts 
       use zero to represent intertoken space */
    p = 0;
    /* finally update character frequencies */
    while( *tok ) {
      q = (unsigned char)*tok;
      learner.dig[p][q]++;
      p = q;
      tok++;
    }
    if( p ) { /* don't count digram 00 */ 
      learner.dig[p][0]++;
    }
  }
}

/* initialize global learner object */
void init_learner() {
    int i, j;

    /* some constants */
    learner.max_tokens = default_max_tokens;
    learner.max_token_bits = default_max_token_bits;
    learner.full_token_count = 0;
    learner.unique_token_count = 0;
    learner.logZ = 0.0;

    /* init character frequencies */
    for(i = 0; i < ASIZE; i++) { 
      for(j = 0; j < ASIZE; j++) { 
	learner.dig[i][j] = 0.0; 
      }
    }

    /* open temporary file for writing tokens */
    learner.tmp = tmpfile();

    /* allocate room for hash */
    learner.hash = calloc(learner.max_tokens, sizeof(l_item));
    if( !learner.hash ) {
	fprintf(stderr, 
		"error: not enough memory? I couldn't allocate %li bytes\n", 
		(sizeof(l_item) * learner.max_tokens));
	exit(0);
    }
}


/* calculates the normalizing constant */
double learner_Z() {
    long int i;
    long int c = 0;
    double t1 = 0.0;
    double t2 = 0.0;

    for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) ) {
	    t1 += exp(learner.hash[i].ref);
	    t2 += exp(learner.hash[i].lam + learner.hash[i].ref);
	    if( ++c >= learner.unique_token_count ) {
		c = 0;
		break;
	    }
	}
    }
    return (1.0 - t1) + t2;
}

/* calculates the divergence but needs normalizing constant */
double learner_divergence(double z) {
    long int i;
    long int c = 0;
    double t = 0.0;

    for(i = 0; i < learner.max_tokens; i++) {
	if( FILLEDP(&learner.hash[i]) ) {
	    t += learner.hash[i].lam * 
	      ((double)learner.hash[i].count/learner.full_token_count);
	    if( ++c >= learner.unique_token_count ) {
		c = 0;
		break;
	    }
	}
    }
    return -log(z) + t;
}

/* tolerance for the error in divergence. */
#define TOL 0.005
/* minimizes the divergence by solving for lambda one 
   component at a time.  */
void minimize_learner_divergence() {
  long int i;
  long int c = 0;
  double d, dd, z, thresh, old_lam;

  if( options & (1<<OPTION_VERBOSE) ) {
    fprintf(stdout, "optimizing model divergence\n");
  }

  z = learner_Z();
  dd = learner_divergence(z);
  do {
    d = dd;     /* save old divergence */

    thresh = 0.0;
    /* change each lambda value one at a time,
       since we only rarely recalculate Z,
       we randomly decide to update a given lambda to
       prevent large errors and periodic behaviour */
    for(i = 0; i < learner.max_tokens; i++) {
      if( FILLEDP(&learner.hash[i]) ) {
	if( rand() > (RAND_MAX>>1)) {
	  old_lam = learner.hash[i].lam;
	  learner.hash[i].lam = log( (z - exp(learner.hash[i].lam + learner.hash[i].ref)) *
				     ((double)learner.hash[i].count / ((double)(learner.full_token_count - learner.hash[i].count))) )
	    - learner.hash[i].ref;
	  thresh += fabs(old_lam - learner.hash[i].lam) * learner.hash[i].ref;
	}
	if( thresh > learner.full_token_count ) {
	  z = learner_Z();
	  thresh = 0.0;
	}
	if( ++c >= learner.unique_token_count ) {
	  c = 0;
	  break;
	}
      }
    }
    /* update values */
    z = learner_Z();
    dd = learner_divergence(z);

    if( options & (1<<OPTION_VERBOSE) ) {
      fprintf(stdout, "divergence change %f --> %f\n",d, dd);
    }

  } while( fabs(d - dd) > TOL );

  /* now save values */
  learner.logZ = log(z);
  learner.divergence = dd;

}

/* initializes to the Laplacian parameters */ 
void init_dirichlet() {
    int i;

    dirichlet.alpha = 1.0;
    for(i = 0; i < ASIZE; i++) {
	dirichlet.u[i] = 1.0/((double)ASIZE);
    }
}

/* calculates the most probable Dirichlet parameters 
   given digram counts. Method described in MacKay and Peto (1995)
   Since we don't count the 00 digram, including the prior will 
   overestimate the 00 transition probability. However, for this 
   transition as well as the others, the Dirichlet will have practically
   no effect when there's enough data in the corpus.
*/
void optimize_dirichlet() {
    int i, j;
    long int k;
    float G[ASIZE];
    float H[ASIZE];
    float V[ASIZE];
    float F[ASIZE];
    double prev_alpha;
    double K, t;

    /* precompute useful constants */
    for(j = 0; j < ASIZE; j++) {
	G[j] = 0.0;
	H[j] = 0.0;
	V[j] = 0.0;
	F[j] = 0.0;
	for(i = 0; i < ASIZE; i++) {
	    for(k = 1; k < learner.dig[i][j]; k++) {
		G[j] += 1.0/k;
		H[j] += 1.0/((double)k*k);
	    }
	    if( learner.dig[i][j] > 0 ) { 
		V[j]++; 
		F[j] += learner.dig[i][j];
	    }
	}
	if( options & (1<<OPTION_DEBUG) ) { 
	    fprintf(stdout, 
		    "symbol %d has G = %g H = %g V = %g F = %g\n", 
		    j, G[j], H[j], V[j], F[j]);
	}
    }

    /* now optimize the dirichlet parameters */
    do {
	prev_alpha = dirichlet.alpha;
	K = 0.0;
	for(i = 0; i < ASIZE; i++) {
	    K += log((F[i] + prev_alpha)/prev_alpha) +
		(0.5) * F[i]/(prev_alpha * (F[i] + prev_alpha));
	}
	dirichlet.alpha = 0.0;
	for(j = 0; j < ASIZE; j++) {
	    t = K - G[j];
	    dirichlet.u[j] = 2 * V[j] /(t + sqrt(t * t + 4 * H[j] * V[j]));
	    dirichlet.alpha += dirichlet.u[j];
	    //fprintf(stderr, "D = %g m[%i] = %g ", t * t - 4 * H[j] * V[j], j, dirichlet.m[j]);
	    //fprintf(stderr, "info: dirichlet alpha %g --> %g\n", prev_alpha, dirichlet.alpha);
	}
	if( options & (1<<OPTION_VERBOSE) ) {
	    fprintf(stdout, 
		    "dirichlet alpha %g --> %g\n", 
		    prev_alpha, dirichlet.alpha);
	}
    } while( fabs(prev_alpha - dirichlet.alpha) > TOL );

    if( options & (1<<OPTION_DEBUG) ) {
	for(j = 0; j < ASIZE; j++) {
	    fprintf(stdout, 
		    "dirichlet u[%d] =  %g\n", 
		    j, dirichlet.u[j]);
	}
    }
}

/* computes the _logarithm_ of digram transition probabilities */
void compute_digram_probabilities() {
    int i,j;
    double t;
    
    for(i = 0; i < ASIZE; i++) {
	t = 0.0;
	for(j = 0; j < ASIZE; j++) {
	    t += learner.dig[i][j];
	}
	for(j = 0; j < ASIZE; j++) {
	    learner.dig[i][j] = 
		log(((double)learner.dig[i][j] + dirichlet.u[j]) / 
		    (t + dirichlet.alpha));
	    if( options & (1<<OPTION_DEBUG) ) {
		fprintf(stdout,
			"learner.dig[%d][%d] = %f\n", 
			i, j, learner.dig[i][j]);
	    }
	}
    }
}

void optimize_learner_and_save() {
    long int i;
    double t;
    unsigned long int id;
    l_item *k;

    char buf[BUFLEN];
    char tok[MAX_TOKEN_LEN];
    
    char *p;
    char *q = tok;
    int pp, pc;

    if(100 * learner.unique_token_count >= HASH_FULL * learner.max_tokens) { 
      fprintf(stderr, 
	      "warning: table full, some tokens ignored - try with option -h %i\n",
	      learner.max_token_bits + 1);
    } else if( learner.unique_token_count <= 0 ) {
      fprintf(stderr, 
	      "warning: no tokens matched - have I learned nothing?\n");
    }

    if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stdout, 
		"picked up %li (%li distinct) tokens\n", 
		learner.full_token_count, learner.unique_token_count);
	fprintf(stdout, 
		"calculating reference word weights\n");
    }

    /* find best Dirichlet parameters */
    init_dirichlet();
    optimize_dirichlet();

    /* estimate the conditional digram probabilities */
    compute_digram_probabilities();

    /* now we calculate the logarithmic word weight
       from the digram model, for each token in the hash */
    rewind(learner.tmp);
    t = 0.0;
    pp = 0;
    while( !feof(learner.tmp) && fgets(buf, BUFLEN, learner.tmp) ) {
	p = buf;
	while( *p ) {
	    if( *p != -1) {
		pc = (unsigned char)*p;
		t += learner.dig[pp][pc];
		*q++ = *p; /* copy into tok */
	    } else { /* interword space */ 
		pc = 0;
		t += learner.dig[pp][pc];
		*q = 0; /* append NUL to tok */

		/* now write weight in hash */
		id = hash(tok, strlen(tok), 0);
		k = find_in_learner(id); /* guaranteed to be found */
		k->ref = t;
	
		if( options & (1<<OPTION_DEBUG) ) {
		    fprintf(stdout, 
			    "%s ref_score %g (random ref_score %g)\n", 
			    tok, t, 
			    -log((double)ASIZE - 1.0) - 
			    (strlen(tok) * log((double)ASIZE)) );
		}
		/* prepare for next token */
		t = 0.0;
		q = tok; /* reset q */
	    }
	    pp = pc;
	    p++;
	}
    }
    /* finish last word */
    t += learner.dig[pp][0];
    *q = 0;
    id = hash(tok, strlen(tok), 0);
    k = find_in_learner(id); /* guaranteed to be found */
    k->ref = t;

    /* done */
    fclose(learner.tmp);

    if( options & (1<<OPTION_VERBOSE) ) {
	t = 0.0;
	for(i=0; i < learner.max_tokens; i++) {
	    if( FILLEDP(&learner.hash[i]) ) {
		t += exp(learner.hash[i].ref);
	    }
	}
	fprintf(stdout, 
		"tokens account for %g of reference probability mass\n", t);
    }

    /* now we have the reference measure and token frequencies
       so we can optimize divergence and calculate normalizing constant */

    if( !(options & (1<<OPTION_REFMODEL))) {
	minimize_learner_divergence();
    }

    /* now save the model to a file */
    save_learner();
}


/***********************************************************
 * MULTIBYTE FILE HANDLING FUNCTIONS                       *
 * this is suitable for any locale whose character set     *
 * encoding doesn't include NUL bytes inside characters    *
 ***********************************************************/

/* returns true if the line should be processed further
   depends on global mbox state */
/* note: mail filtering is very complicated, since
   any message can be mangled and requoted by a sender
   this will only cope with clean mail boxes */
int mbox_line_filter(char *line) {

  int process_line = 0; /* by default we skip the line */

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case UNDEF:
    if( mbox.prev_line_empty && 
	(!strncmp(line, "From ", 5) || 
	 ((line[0] == '-') && (line[1] == '-') && !iswspace(line[2]))) ) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    }
    break;
  case HEADER:
    if( *line == '\n' ) {
      mbox.state = BODY;
    } else if( !strncmp(line, "Content-Type:", 13) ) {
      mbox.checked_content_type = 1;
      if( strstr(line + 13, "text/") || 
	  strstr(line + 13, "TEXT/") ) {
	mbox.mime_type = 1; /* good */
      } else {
	mbox.mime_type = 0; /* bad */
      }
    } else if( !strncmp(line, "Subject:", 8) ||
	       !(strncmp(line, "From:", 5)) ) {
      /* process subject and from line like body */
      process_line = 1;
    }
    break;
  case BODY:
      if( mbox.prev_line_empty && 
	  (!strncmp(line, "From ", 5) || 
	   ((line[0] == '-') && (line[1] == '-') && !isspace(line[2])))) {
	/* if it starts with From_ or else it looks like --xxxxx, 
	   it signals a new header */
	mbox.state = HEADER;
	mbox.checked_content_type = 0;
      } else if( mbox.checked_content_type && mbox.mime_type ) { 
	process_line = 1;
      } else if( !mbox.checked_content_type ) {
	process_line = 1; /* be lenient */
      }
      break;
    }
    mbox.prev_line_empty = (*line == '\n') ? 1 : 0; /* for next time */

    return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void xml_character_filter(char *line) {

  char *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == '<' ) {
	if( !strncmp(line + 1, "!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( !strncasecmp(line + 1, "script", 6) ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( isalpha(line[1]) || ispunct(line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( (line[1] == '>') && 
	  (isalpha(*line) || ispunct(*line)) ) {
	xml.state = TEXT;
	line++; 
      }
      break;
    case COMMENT:
      if( line[0] == '-' && !strncmp(line + 1, "->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == '<' ) {
	if( !strncasecmp(line + 1, "/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = '\0'; /* mark the end of the clean text string */
}

#define MULTIBYTE_EPSILON 10 /* enough for a multibyte char and a null char */

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void process_file(FILE *input, int (*line_filter)(),
		  void (*character_filter)(), 
		  void (*word_fun)(), void (*line_fun)()) {
  char *s;
  int i, k, l;

#if defined HAVE_LIBBOOST_REGEX
  int wclen;
  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  int tok_len;
#endif

  char *p, *q;
  char tok[MAX_TOKEN_LEN];

  regmatch_t pmatch[MAX_SUBMATCH];
  int eflag, z, j;

  while( !feof(input) ) {
    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* next we check to see if this line should be skipped */
    if( *textbuf && (!line_filter || (*line_filter)(textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      /* damn - we compiled wide character regexes, 
	 now we've got to use them (not really, but who in their right
	 mind would compile with boost if they're not going to convert
	 to wide characters anyway) */

      /* convert as much as we can of the line into wide characters */
      s = textbuf;
      k = textbuf_len;
      wp = wc_textbuf;
      wclen = 0;
      /* since we ensured textbuf_len <= wctextbuf_len
	 there will never be overflow of wctextbuf below */
      while( k > 0 ) {
	l = mbrtowc(wp, s, k, &input_shiftstate);
	if( l > 0 ) {
	  wp++;
	  wclen++;
	  k -= l;
	  s += l;
	} else if( l == 0 ) {
	  break;
	} else if( l == -1 ) {
	  /* try to be robust */
	  s++; 
	  k--;
	  memset(&input_shiftstate, 0, sizeof(mbstate_t));
	} else if( l == -2) {
	  /* couldn't parse a complete character */
	  break;
	}
      }
      *wp = L'\0';

      /* now process the regex */
      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(z = 1; (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
 	  (*word_fun)(tok); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect an ordinary string */
      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
 	  for(z = 1; (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = textbuf[k + j];
		} else {
		  *q++ = tolower(textbuf[k + j]);
		}
 	      }
 	    }
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
 	  (*word_fun)(tok); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens and passes them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	p = textbuf;
	q = tok;
	while( *p ) {
	  if( isalpha(*p) ) {
	    if( q < tok + MAX_TOKEN_LEN ) {
	      if( !(options & (1<<OPTION_CASEN)) ) {
		*q++ = tolower(*p);
	      } else {
		*q++ = *p;
	      }
	    }
	  } else { /* token boundary */
	    *q = '\0';
	    q = tok;
	    /* let each category process the token */
	    (*word_fun)(q);
	  }
	  p++;
	}
      }

      /* now summarize this line if required */
      if( line_fun ) { (*line_fun)(); }
      
    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }
  }
}

/***********************************************************
 * WIDE CHARACTER FILE HANDLING FUNCTIONS                  *
 * this is needed for any locale whose character set       *
 * encoding can include NUL bytes inside characters        *
 ***********************************************************/

/* returns true if the line should be processed further
   depends on global mbox state */
int w_mbox_line_filter(wchar_t *line) {

  int process_line = 0; /* by default we skip the line */

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case UNDEF:
    if( mbox.prev_line_empty && 
	(!wcsncmp(line, L"From ", 5) || 
	 ((line[0] == L'-') && (line[1] == L'-') && !iswspace(line[2]))) ) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    }
    break;
  case HEADER:
    if( *line == L'\n' ) {
      mbox.state = BODY;
    } else if( !wcsncmp(line, L"Content-Type:", 13) ) {
      mbox.checked_content_type = 1;
      if( wcsstr(line + 13, L"text/") || 
	  wcsstr(line + 13, L"TEXT/") ) {
	mbox.mime_type = 1; /* good */
      } else {
	mbox.mime_type = 0; /* bad */
      }
    } else if( !wcsncmp(line, L"Subject:", 8) ||
	       !(wcsncmp(line, L"From:", 5)) ) {
      /* process subject and from line like body */
      process_line = 1;
    }
    break;
  case BODY:
    if( mbox.prev_line_empty && 
	(!wcsncmp(line, L"From ", 5) || 
	 ((line[0] == L'-') && (line[1] == L'-') && !iswspace(line[2])))) {
      /* if it starts with From_ or else it looks like --xxxxx, 
	 it signals a new header */
      mbox.state = HEADER;
      mbox.checked_content_type = 0;
    } else if( mbox.checked_content_type && mbox.mime_type ) { 
      process_line = 1;
    } else if( !mbox.checked_content_type ) {
      /* unless we explicitly are told it's not text, we process it */
      /* this will let quoted uuencoded content through :-( */
      process_line = 1;
    }
    break;
  }
  mbox.prev_line_empty = (*line == L'\n') ? 1 : 0; /* for next time */

  return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void w_xml_character_filter(wchar_t *line) {

  wchar_t *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == L'<' ) {
	if( !wcsncmp(line + 1, L"!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( wcsncasecmp(line + 1, L"script", 6) != 0 ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( iswalpha(line[1]) || iswpunct(line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( (line[1] == L'>') && 
	  (iswalpha(*line) || iswpunct(*line)) ) {
	xml.state = TEXT;
	line++; 
      }
      break;
    case COMMENT:
      if( line[0] == L'-' && !wcsncmp(line + 1, L"->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == L'<' ) {
	if( !wcsncasecmp(line + 1, L"/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = L'\0'; /* mark the end of the clean text string */
}

#define MULTIBYTE_EPSILON 10 /* enough for a multibyte char and a null char */

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void w_process_file(FILE *input, int (*line_filter)(),
		    void (*character_filter)(), 
		    void (*word_fun)(), void (*line_fun)()) {
  char *s;
  int i, k, l;
  int wclen;

  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  int tok_len;
  char *q;
  char tok[MAX_TOKEN_LEN];

  regmatch_t pmatch[MAX_SUBMATCH];
  int eflag, z, j;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));

  while( !feof(input) ) {

    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* now convert the line into a wide character string */
    if( textbuf_len > wc_textbuf_len ) {
      wc_textbuf_len = textbuf_len;
      wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
      if( !wc_textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for wide character conversion (%d bytes)\n",
		wc_textbuf_len * sizeof(wchar_t));
	exit(0);
      }
    }

    /* convert as much as we can of the line into wide characters */
    s = textbuf;
    k = textbuf_len;
    wp = wc_textbuf;
    wclen = 0;
    /* since we ensured textbuf_len <= wctextbuf_len
       there will never be overflow of wctextbuf below */
    while( k > 0 ) {
      l = mbrtowc(wp, s, k, &input_shiftstate);
      if( l > 0 ) {
	wp++;
	wclen++;
	k -= l;
	s += l;
      } else if( l == 0 ) {
	break;
      } else if( l == -1 ) {
	/* try to be robust */
	s++; 
	k--;
	memset(&input_shiftstate, 0, sizeof(mbstate_t));
      } else if( l == -2) {
	/* couldn't parse a complete character */
	break;
      }
    }
    *wp = L'\0';

    /* next we check to see if this line should be skipped */
    if( (wclen > 0) && 
	(!line_filter || (*line_filter)(wc_textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(wc_textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(z = 1; (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
 	  (*word_fun)(tok); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect a multibyte string */
      if( textbuf_len > aux_textbuf_len ) {
	aux_textbuf_len = textbuf_len;
	aux_textbuf = realloc(aux_textbuf, aux_textbuf_len);
	if( !aux_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for auxiliary text buffer (%d bytes)\n",
		  aux_textbuf_len);
	  exit(0);
	}
      }
      l = wcstombs(aux_textbuf, wc_textbuf, aux_textbuf_len - 1);
      aux_textbuf[aux_textbuf_len - 1] = '\0';

      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, aux_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
 	  for(z = 1; (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = aux_textbuf[k + j];
		} else {
		  *q++ = tolower(aux_textbuf[k + j]);
		}
 	      }
 	    }
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
 	  (*word_fun)(tok); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens, converting them
	 to multibyte representation before passing them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	wp = wc_textbuf;
	q = tok;
	memset(&tok_shiftstate, 0, sizeof(mbstate_t));
	while( *wp ) {
	  if( iswalpha(*wp) ) {
	    if( !(options & (1<<OPTION_CASEN)) ) {
	      *wp = towlower(*wp);
	    }

	    if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
	      tok_len = wcrtomb(q, *wp, &tok_shiftstate); 
	      if( (tok_len > -1) ) {
		q += tok_len;
	      }
	    }
	  } else { /* token boundary */
	    *q = '\0';
	    q = tok;
	    /* let each category process the token */
	    (*word_fun)(q);
	  }
	  wp++;
	}
      }

      /* now summarize this line if required */
      if( line_fun ) { (*line_fun)(); }
      
    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }

  }
}

/***********************************************************
 * MAIN FUNCTIONS                                          *
 ***********************************************************/

char *sanitize_path(char *in) {
  char *q;
  char *path;
  int l;

  if( (*in != '/') && (*in != '.') && (path = getenv("DBACL_PATH")) ) {
    l = strlen(path);
    q = malloc(l + strlen(in) + 3);
    strcpy(q, path);
    if( q[l - 1] != '/' ) {
      q[l] = '/'; 
      q[l + 1] = 0;
    }
    strcat(q, in);
    return q;
  } else {
    return in;
  }
}

int main(int argc, char **argv) {

    int k;
    FILE *input;
    char op;

#if defined HAVE_LIBBOOST_REGEX
    wchar_t *rstring;
#endif

    void (*preprocess_fun)() = NULL;
    void (*word_fun)() = NULL;
    void (*line_fun)() = NULL;
    void (*postprocess_fun)() = NULL;

    /* set up internationalization */
    if( !setlocale(LC_ALL, "") ) {
      fprintf(stderr, 
	      "warning: could not set locale, internationalization disabled\n");
    } else {
      if( options & (1<<OPTION_VERBOSE) ) {
	fprintf(stderr, 
		"warning: international locales not supported\n");
      }
    }

    /* parse the options */
    while( (op = getopt(argc, argv, "DRVNivnrh:c:l:T:f:g:j:")) > -1 ) {

      switch(op) {
      case 'D':
	options |= (1<<OPTION_DEBUG);
	break;
      case 'N':
	options |= (1<<OPTION_POSTERIOR);
	break;
      case 'R':
	if( cat_count >= MAX_CAT ) {
	  fprintf(stderr, 
		  "warning: maximum reached, random text category omitted\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, 
		  "error: cannot use options -l and -R together\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_CLASSIFY);

	  cat[cat_count].filename = "random";

	  init_category(&cat[cat_count]);
	  init_purely_random_text_category(&cat[cat_count]);

	  cat_count++;
	}
	break;
      case 'T':
	if( !strncasecmp(optarg, "email", 5) ) {
	  options |= (1<<OPTION_MBOX_FORMAT);
	  options |= (1<<OPTION_XML); /* filter out HTML from messages */
	} else if( !strncasecmp(optarg, "xml", 3) ) {
	  options |= (1<<OPTION_XML);
	} else { /* default */
	  options |= (1<<OPTION_TEXT_FORMAT);
	}
	break;
      case 'V':
	fprintf(stdout, "dbacl version %s\n", VERSION);
#if defined HAVE_LIBBOOST_REGEX
	fprintf(stdout, "Using BOOST wide character EXTENDED regexes.\n");
#else
	fprintf(stdout, "Using GNU (glibc) EXTENDED regexes.\n");
#endif
	exit(1);
	break;
      case 'h': /* select memory size in megabytes */
	default_max_token_bits = strtod(optarg, NULL);
	if( default_max_token_bits > MAX_TOKEN_BITS ) {
	  fprintf(stderr, 
		  "warning: maximum hash size will be 2^%d\n", 
		  MAX_TOKEN_BITS);
	  default_max_token_bits = MAX_TOKEN_BITS;
	}
	default_max_tokens = (1<<default_max_token_bits);
	break;
      case 'j':
	options |= (1<<OPTION_CASEN);
	break;
      case 'n':
	options |= (1<<OPTION_SCORES);
	break;
      case 'c':
	if( cat_count >= MAX_CAT ) {
	  fprintf(stderr, 
		  "warning: maximum reached, category ignored\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, 
		  "error: cannot use options -l and -c together\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_CLASSIFY);

	  cat[cat_count].filename = sanitize_path(optarg);

	  if( !*optarg ) {
	    fprintf(stderr, "error: category needs a name\n");
	    usage(argv);
	    exit(0);
	  }
	  if( load_category(&cat[cat_count]) ) {
	    init_category(&cat[cat_count]);
	  } else {
	    fprintf(stderr, 
		    "error: could not load category %s\n", 
		    cat[cat_count].filename);
	    exit(0);
	  }
	  cat_count++;
	}
	break;
      case 'r':
	options |= (1<<OPTION_REFMODEL);
	break;
      case 'f':
	if( filter_count >= MAX_CAT ) {
	  fprintf(stderr, "warning: maximum reached, filter ignored\n");
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, "error: cannot use options -l and -f together\n");
	  usage(argv);
	  exit(0);
	} else if( !*optarg ) {
	  fprintf(stderr, "error: filter must be category name or number\n");
	  usage(argv);
	  exit(0);
	} else {
	  options |= (1<<OPTION_FILTER);
	  filter[filter_count] = -1;

	  /* see if it's a known category */
	  for(k = 0; k < cat_count; k++) {
	    if( !strcmp(cat[k].filename, optarg) ) {
	      filter[filter_count] = k;
	      break;
	    }
	  }
	  /* if not recognized, see if it's a number */
	  if( filter[filter_count] < 0 ) {
	    filter[filter_count] = strtod(optarg, NULL) - 1;
	  }
	  if( filter[filter_count] < 0 ) { /* still not recognized */
	    fprintf(stderr, 
		    "error: unrecognized category in -f option [%s]\n", 
		    optarg);
	    usage(argv);
	    exit(0);
	  }
	  filter_count++;
	}
	break;
      case 'v':
	options |= (1<<OPTION_VERBOSE);
	break;
      case 'l': 
	if( options & (1<<OPTION_CLASSIFY) ) {
	  fprintf(stderr, 
		  "error: cannot use options -l and -c together\n");
	  usage(argv);
	  exit(0);
	} else if( options & (1<<OPTION_LEARN) ) {
	  fprintf(stderr, 
		  "error: option -l can only occur once\n");
	  exit(0);
	} else {
	  options |= (1<<OPTION_LEARN);
	  learner.filename = sanitize_path(optarg);
	  if( !*learner.filename ) {
	    fprintf(stderr, 
		    "error: category needs a name\n");
	    usage(argv);
	    exit(0);
	  }
	}
	break;
      case 'g':
	if( options & (1<<OPTION_CLASSIFY) ) {
	  fprintf(stderr, "error: must use option -l together with -F\n");
	  usage(argv);
	  exit(0);
	} else {
#if defined HAVE_LIBBOOST_REGEX
	  /* BOOST regexes must be wide */
	  if( !(rstring = malloc((strlen(optarg) + 1) * sizeof(wchar_t))) ) {
	    fprintf(stderr, 
		    "warning: could not prepare regular expression '%s', ignored\n",
		    optarg);
	  } else if( mbstowcs(rstring, optarg, strlen(optarg)) < 0 ) {
	    fprintf(stderr,
		    "warning: could not convert regular expression '%s', ignored\n",
		    optarg);
	  } else {
	    rstring[strlen(optarg)] = L'\0';
	    if( regcomp(&re[regex_count].regex, rstring, REG_EXTENDED) != 0 ) {
	      fprintf(stderr, 
		      "warning: could not compile regular expression '%s', ignored\n", 
		      optarg);
	    } else {
	      re[regex_count].string = rstring;
	      regex_count++;
	    }
	  }
#else
	  /* GNU regexes use regular strings */
	  if( regcomp(&re[regex_count].regex, optarg, REG_EXTENDED) != 0 ) {
	    fprintf(stderr, 
		    "warning: could not compile regular expression '%s', ignored\n", 
		    optarg);
	  } else {
	    re[regex_count].string = optarg;
	    regex_count++;
	  }
#endif
	}
	break;
      case 'i':
	  options |= (1<<OPTION_I18N);
	  if( !strcmp(nl_langinfo(CODESET), "UTF-8") ) {
	    fprintf(stderr, "warning: you have UTF-8, so -i is not needed.\n");
	  }
	break;
      default:
	break;
      }
    }
    
    /* consistency check */
    if( ((options>>OPTION_CLASSIFY) & 1) + 
	((options>>OPTION_LEARN) & 1) != 1 ) {
      fprintf(stderr, 
	      "error: please use either -c or -l option\n");
      usage(argv);
      exit(0);
    }

    if( ((options>>OPTION_TEXT_FORMAT) & 1) + 
	((options>>OPTION_MBOX_FORMAT) & 1) > 1 ) {
      fprintf(stderr, 
	      "error: please use one of either -T text or -T email options\n");
      usage(argv);
      exit(0);
    }

    if( !regex_count ) {
      options |= (1<<OPTION_NOREGEX);
    } else if( options & (1<<OPTION_I18N) ) {
#if defined HAVE_LIBBOOST_REGEX
#else
	fprintf(stderr, 
		"warning: regexes operate in multibyte encoding.");
#endif	
    }

    /* set up callbacks */
    if( options & (1<<OPTION_CLASSIFY) ) {
      if( cat_count == 1 ) {
	/* single category */

	init_empirical(); /* sets cached to zero */
	    
	preprocess_fun = NULL;
	word_fun = score_word;
	if( options & (1<<OPTION_FILTER) ) {
	  empirical.cached = 1;
	  line_fun = line_score_single_category;
	  postprocess_fun = NULL;
	} else {
	  line_fun = NULL;
	  postprocess_fun = score_single_category;
	}

      } else {
	/* multiple categories */

	/* we don't any other part of empirical */
	empirical.full_token_count = 0; 

	preprocess_fun = NULL;
	word_fun = score_word;
	if( options & (1<<OPTION_FILTER) ) {
	  empirical.cached = 1;
	  line_fun = line_score_multiple_categories;
	  postprocess_fun = NULL;
	} else {
	  line_fun = NULL;
	  postprocess_fun = score_multiple_categories;
	}

      }
    } else if( options & (1<<OPTION_LEARN) ) {
      /* category learning */

      preprocess_fun = init_learner;
      word_fun = hash_word_and_learn;
      line_fun = NULL;
      postprocess_fun = optimize_learner_and_save;

    } else { /* something wrong ? */
      usage(argv);
      exit(0);
    }

    if( preprocess_fun ) { (*preprocess_fun)(); }

    /* preallocate primary text holding buffer */
    textbuf_len = BUFLEN;
    textbuf = malloc(textbuf_len);

    /* now process each file on the command line,
       or if none provided read stdin */
    while( (optind > -1) && *(argv + optind) ) {
      /* if it's a filename, process it */
      if( (input = fopen(argv[optind], "r")) ) {

	options |= (1<<INPUT_FROM_CMDLINE);

	if( (options & (1<<OPTION_VERBOSE)) && 
	    !(options & (1<<OPTION_CLASSIFY))) {
	  fprintf(stdout, "processing file %s\n", argv[optind]);
	}

	/* set some initial options */
	xml.state = TEXT; 
	if( options & (1<<OPTION_MBOX_FORMAT) ) {

	  mbox.state = UNDEF;
	  mbox.prev_line_empty = 1;
	  if( !(options & (1<<OPTION_I18N)) ) {
	    process_file(input, mbox_line_filter, 
			 ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
			 word_fun, line_fun);
	  } else {
	    w_process_file(input, w_mbox_line_filter, 
			   ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			   word_fun, line_fun);
	  }
	} else { 
	  /* default to text */
	  if( !(options & (1<<OPTION_I18N)) ) {
	    process_file(input, NULL, 
			 ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
			 word_fun, line_fun);
	  } else {
	    w_process_file(input, NULL, 
			   ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			   word_fun, line_fun);
	  }
	}

	fclose(input);

      } else { /* unrecognized file name */

	fprintf(stderr, "error: couldn't open %s\n", argv[optind]);
	usage(argv);
	exit(0);

      }
      optind++;
    }
    /* in case no files were specified, get input from stdin */
    if( !(options & (1<<INPUT_FROM_CMDLINE)) ) {

      if( (options & (1<<OPTION_VERBOSE)) && 
	  !(options & (1<<OPTION_CLASSIFY)) ) {
	fprintf(stdout, "taking input from stdin\n");
      }

      /* set some initial options */
      xml.state = TEXT; 

      if( options & (1<<OPTION_MBOX_FORMAT) ) {

	mbox.state = UNDEF;
	mbox.prev_line_empty = 1;
	if( !(options & (1<<OPTION_I18N)) ) {
	  process_file(stdin, mbox_line_filter, 
		       ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		       word_fun, line_fun);
	} else {
	  w_process_file(stdin, w_mbox_line_filter, 
			 ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			 word_fun, line_fun);
	}

      } else { 
	/* default to text */
	if( !(options & (1<<OPTION_I18N)) ) {
	  process_file(stdin, NULL,
		       ((options & (1<<OPTION_XML)) ? xml_character_filter : NULL), 
		       word_fun, line_fun);	
	} else {
	  w_process_file(stdin, NULL, 
			 ((options & (1<<OPTION_XML)) ? w_xml_character_filter : NULL), 
			 word_fun, line_fun);
	}
      }

    }
  
    if( postprocess_fun ) { (*postprocess_fun)(); }

    /* free some global resources */
    free(textbuf);

    for(k = 0; k < regex_count; k++) {
      regfree(&re[k].regex);
    }

    exit(exit_code);
}
